package de.uni_marburg.powersort.sort;

import java.util.*;
import java.util.stream.Collectors;

public class ASort {

    private static class Run {
        int start, end, power;

        Run(int start, int end, int power) {
            this.start = start;
            this.end = end;
            this.power = power;
        }
    }

    public static <T> void sort(T[] array, Comparator<? super T> comparator) {
        if (array == null || comparator == null) {
            throw new IllegalArgumentException("Array and comparator must not be null");
        }

        int n = array.length;
        if (n < 2) return;

        Deque<Run> runs = new ArrayDeque<>();
        int i = 0;

        while (i < n) {
            int j = extendRunRight(array, comparator, i, n);
            int power = calculatePower(runs, i, j);

            while (!runs.isEmpty() && power <= runs.peekLast().power) {
                mergeRuns(array, comparator, runs);
            }

            runs.addLast(new Run(i, j, power));
            i = j;
        }

        while (runs.size() > 1) {
            mergeRuns(array, comparator, runs);
        }
    }

    private static <T> int extendRunRight(T[] array, Comparator<? super T> comparator, int start, int end) {
        int runEnd = start + 1;

        if (runEnd == end) return runEnd;

        if (comparator.compare(array[runEnd], array[start]) < 0) {
            while (runEnd < end && comparator.compare(array[runEnd], array[runEnd - 1]) < 0) {
                runEnd++;
            }
            reverseRange(array, start, runEnd);
        } else {
            while (runEnd < end && comparator.compare(array[runEnd], array[runEnd - 1]) >= 0) {
                runEnd++;
            }
        }
        return runEnd;
    }

    private static <T> void reverseRange(T[] array, int start, int end) {
        end--;
        while (start < end) {
            T temp = array[start];
            array[start++] = array[end];
            array[end--] = temp;
        }
    }

    private static int calculatePower(Deque<Run> runs, int currentStart, int currentEnd) {
        if (runs.isEmpty()) return 0;

        Run topRun = runs.peekLast();
        int size1 = topRun.end - topRun.start;
        int size2 = currentEnd - currentStart;

        return Math.max(topRun.power, Integer.highestOneBit(size1 + size2)) + 1;
    }

    private static <T> void mergeRuns(T[] array, Comparator<? super T> comparator, Deque<Run> runs) {
        if (runs.size() < 2) return;

        Run run2 = runs.removeLast();
        Run run1 = runs.removeLast();

        merge(array, comparator, run1.start, run1.end, run2.end);
        int newPower = Math.max(run1.power, run2.power) + 1;

        runs.addLast(new Run(run1.start, run2.end, newPower));
    }

    private static <T> void merge(T[] array, Comparator<? super T> comparator, int lo, int mid, int hi) {
        List<T> merged = Arrays.stream(array, lo, mid)
                .sorted(comparator)
                .toList();

        int i = 0, j = mid, k = lo;

        while (i < merged.size() && j < hi) {
            array[k++] = comparator.compare(merged.get(i), array[j]) <= 0 ? merged.get(i++) : array[j++];
        }

        while (i < merged.size()) {
            array[k++] = merged.get(i++);
        }
    }

    public static void main(String[] args) {
        Integer[] array = {0, 1, 2, 3, 4, 5, -5, -4, -3, -2, -1, -10, -9, -8, -7, -6};
        System.out.println("Before sorting:");
        Arrays.stream(array).forEach(e -> System.out.print(e + " "));

        sort(array, Integer::compareTo);

        System.out.println("\nAfter sorting:");
        Arrays.stream(array).forEach(e -> System.out.print(e + " "));
    }
}




